"""
Remove a list of email templates from SES
"""
from aws_teams_logger.log import LOG
from .utils import get_templates, setup_logging
from ..ses import SESHelper


def delete_templates(profile_name=None):
    setup_logging()
    ses = SESHelper(profile_name=profile_name)

    templates = [template.name for template in get_templates()]

    LOG.info('Attempting to delete %d templates...', len(templates))

    for template in templates:
        LOG.info('[ %s ]', template)
        success = ses.delete_template(template)
        if success:
            LOG.info('Deleted')
        else:
            LOG.error('There was an issue deleting the template')



import os

__all__ = ['LOG_CFG', 'AWS_ACCOUNT_NAME',
           'TEAMS_LOG_LVL', 'TEAMS_EMAIL',
           'SES_IDENTITY', 'DEV_EMAILS',
           'SOURCE_CODE', 'AWS_LOG_GROUP',
           'LOCAL_TZ']


# Minimum level for library logs, to show up in CloudWatch
LOG_CFG = os.getenv('LOG_CFG')

# AWS Account Alias (name), can be optionally set in the environment
#
# If defined, will be used instead of making a call to `iam:ListAccountAliases`
# to retrieve the alias of the current AWS account.
AWS_ACCOUNT_NAME = os.getenv('AWS_ACCOUNT_NAME')

# Minimum log level for messages sent to Teams
TEAMS_LOG_LVL = os.getenv('TEAMS_LOG_LVL')

# MS Teams email
TEAMS_EMAIL = os.getenv('TEAMS_EMAIL')

# SES outbound email address
SES_IDENTITY = os.getenv('SES_IDENTITY')

# Comma delimited field, if provided will send stylized HTML to them
#
# Example:
#   'user1@my.domain.org,user2@my.domain.org'
DEV_EMAILS = os.getenv('DEV_EMAILS')

# Optional link to source code repo for the project
SOURCE_CODE = os.getenv('SOURCE_CODE')

# (ECS Tasks) Optional link to the AWS log group
AWS_LOG_GROUP = os.getenv('AWS_LOG_GROUP')

# Local time zone
LOCAL_TZ = os.getenv('LOCAL_TZ', 'US/Eastern')



from collections import defaultdict

from boto3 import Session, client
from botocore.client import BaseClient
from botocore.exceptions import UnknownServiceError


class ClientCache:

    # boto3 service name
    SERVICE_NAME = None

    # sub-classes can specify that client objects should be thread-safe
    THREAD_SAFE = False

    _clients = defaultdict(dict)

    def __init__(self, region_name='us-east-1', profile_name=None,
                 init_client=False):
        """Initializes a :class:`ClientCache` object.

        `region_name` is the specified region that is bound to a client,
        which defaults to "us-east-1".

        `init_client` will automatically initialize the client for the region,
        and is useful when S3 requests will be made in multiple threads and
        it is desirable to reuse the same client between threads.

        """
        self.region_name = region_name.lower()
        self.profile_name = profile_name

        if init_client:
            _ = self.client

    @property
    def client(self):
        return self._get_client(self.region_name)

    @client.setter
    def client(self, value):
        raise Exception('Member read-only')

    def _get_client(self, region_name):
        """
        Internal method to return a low-level SecretManager client for a given region name
        """
        if region_name not in self._clients[self.SERVICE_NAME]:
            self._clients[self.SERVICE_NAME][region_name] = self._create_client()

        return self._clients[self.SERVICE_NAME][region_name]

    def _create_client(self) -> BaseClient:
        if not self.SERVICE_NAME:
            raise ValueError(
                'Sub-classes must provide a value for "SERVICE_NAME"')

        try:
            if self.THREAD_SAFE or self.profile_name:
                client_func = Session(profile_name=self.profile_name).client
            else:
                client_func = client
            return client_func(self.SERVICE_NAME, self.region_name)

        except UnknownServiceError:
            raise NotImplementedError(
                'Sub-classes must override this method')



import os
from json import dumps
from typing import Union, List, Dict, Any

from aws_teams_logger.log import LOG
from ..client_cache import ClientCache
from ...decorators import log_time
from ...types import as_list


class SESHelper(ClientCache):

    SERVICE_NAME = 'ses'

    @staticmethod
    def _op_is_success(response: Dict[str, Any]) -> bool:
        """Check if response indicates an operation was a success"""
        return response.get('ResponseMetadata', {}).get('HTTPStatusCode') == 200

    @log_time
    def send_templated_email(self,
                             name: str, data: Dict[str, Any],
                             to_addresses: Union[List[str], str],
                             from_address: str,
                             reply_to_addresses: Union[List[str], str]):
        """
        Send templated email via SES client
        """
        response = self.client.send_templated_email(
            Source=from_address,
            Destination={'ToAddresses': as_list(to_addresses)},
            ReplyToAddresses=as_list(reply_to_addresses),
            Template=name,
            TemplateData=dumps(data)
        )

        return response

    def replace_template(self, template_data: Dict[str, Any]) -> bool:
        """
        Update or create an SES template
        """
        try:
            return self.update_template(template_data)

        except self.client.exceptions.TemplateDoesNotExistException:
            LOG.debug('%s: creating template, as it does not exist',
                      template_data['TemplateName'])
            return self.create_template(template_data)

        except self.client.exceptions.InvalidTemplateException as e:
            LOG.warning(e)
            return False

    @log_time
    def update_template(self, template_data: Dict[str, Any]) -> bool:
        """Update an SES template"""
        response = self.client.update_template(
            Template=template_data
        )

        return self._op_is_success(response)

    @log_time
    def create_template(self, template_data: Dict[str, Any]) -> bool:
        """Create a new SES template"""
        response = self.client.create_template(
            Template=template_data
        )

        return self._op_is_success(response)

    @log_time
    def delete_template(self, template_name: str) -> bool:
        """Delete an SES template"""

        response = self.client.delete_template(
            TemplateName=template_name
        )

        return self._op_is_success(response)

    @log_time
    def get_template(self, name: str) -> bool:
        """Get an SES template"""
        response = self.client.get_template(
            TemplateName=name
        )

        return response

    @log_time
    def test_render_template(self, name: str, data: Dict[str, Any],
                             html_file_name='email_body.html') -> bool:
        """
        Attempt to render an SES template using provided template arguments.

        The expected HTML body is rendered in a web browser which will
        automatically be opened.

        """
        import inspect
        import webbrowser

        # Skip Stack @index=1 as that leads to decorators.py
        caller_dir = os.path.dirname((inspect.stack()[2])[1])

        response = self.client.test_render_template(
            TemplateName=name,
            TemplateData=dumps(data)
        )

        render_data = response['RenderedTemplate'].replace('\r\n', '\n')

        header_data, html_data = render_data.split('\n\n', 1)
        LOG.info('Email Headers:\n%s', header_data)

        html_file_path = os.path.join(caller_dir, html_file_name)

        with open(html_file_path, 'w') as out_file:
            out_file.write(html_data)

        LOG.info(f'Wrote HTML file out to: %s', html_file_path)

        webbrowser.open_new_tab(f'file://{html_file_path}')

        return render_data



"""
Project-specific decorators
"""
import functools
from datetime import timedelta
from logging import Logger
from typing import Optional

from .time_util import preferred_clock
from ..log import LOG


def record_time(f=None, log_message: str = None, log_func=print):
    """
    Decorates a function to also log its execution time when it completes.

    `log_message` is a message that can be specified to replace the default
    one, that logs in the format "Successfully ran task {func_name}"

    `log_func` is the function to send the message to, by default it passes
    the message to the 'print' function.

    """
    def decorator(func):

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            start = preferred_clock()

            result = func(*args, **kwargs)

            diff_sec = preferred_clock() - start
            elapsed = timedelta(seconds=diff_sec)

            message = log_message or f'Successfully ran task `{func.__name__.lstrip("_")}`'

            log_func(f'[{elapsed}] {message}')

            return result

        return new_func

    if f:
        return decorator(f)

    return decorator


def log_time(f=None, logger: Logger = LOG, log_method='info',
             log_message: Optional[str] = None):
    """
    Decorator to log the execution time using a `logger` object instead

    :param logger: Optional Logger object, defaults to the root library logger
      if not provided.
    :param log_method: Log method to send messages to, defaults to the 'info'
      method if not provided.
    :param log_message: a message that can be specified to replace the default
      one, that logs in the format "Successfully ran task {func_name}"

    """
    return record_time(f, log_func=getattr(logger, log_method), log_message=log_message)



# -*- coding: utf-8 -*-

"""
MS Teams / Email Logger for AWS
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

AWS Teams Logger is a Python library that forwards errors (failures) and log
messages to a MS Teams channel and an optional list of Developer Emails who may
need to be notified -- adds HTML formatting originally designed for MS Outlook.

Simple Usage for an AWS Lambda function:

    >>> from aws_teams_logger import LambdaLogger
    >>> from logging import getLogger
    >>>
    >>> log = getLogger()
    >>>
    >>> # Note: this is a simplified example, and assumes you define the required
    >>> # environment variables. Otherwise, you'd need to pass the parameters
    >>> # to the decorator class like `@LambdaLogger(teams_email='my-teams-email')
    >>> # in this case.
    >>>
    >>> @LambdaLogger
    >>> def my_lambda_handler(event, context):
    >>>   # This message can be sent to Teams, depending on the enabled log lvl
    >>>   log.info('Hello world!')
    >>>   try:
    >>>     result = 1 / 0
    >>>   except ZeroDivisionError:
    >>>     # This will forward the error to Teams, and notify any Devs via email
    >>>     log.error('Unable to divide by zero', exc_info=True)
    >>>     # Be sure not to re-raise the exception as below, as that will log
    >>>     # a duplicate message to Teams and Outlook
    >>>     # raise


Please see the docs for additional examples and some important how-to's.

"""
__all__ = [
    'LambdaLogger',
    'TaskLogger',
    'set_account_name',
    'upload_templates',
    'delete_templates'
]

import logging

from .loggers import *
from .utils.aws.ses.templates import *

# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
logging.getLogger('aws_teams_logger').addHandler(logging.NullHandler())



"""
High-level implementation of the boto3 library for SES

Read the Docs here:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ses.html

"""

__all__ = ['SESHelper']


from .ses import SESHelper



import functools
import logging
import logging.config
import os
from logging import getLogger, Logger, LoggerAdapter

from .constants import LOG_CFG


# Contains the original logger methods, by default for the `logging` module
#
# Note: don't directly modify this object
_ORIGINAL_LOG_METHODS = {}


def setup_logging():
    """
    Allow for easy insight into SDK behavior.
    """
    if LOG_CFG is not None:
        if os.path.exists(LOG_CFG):
            import json
            with open(LOG_CFG, 'rt') as config_file:
                config = json.load(config_file)
            logging.config.dictConfig(config)
        else:
            level_name = logging._nameToLevel.get(LOG_CFG.upper())
            if level_name:
                logging.basicConfig(level=level_name)
    # we will do most of the logging here, so turn down the requests and botocore libraries
    for lib in 'requests', 'urllib3', 'botocore':
        logging.getLogger(lib).setLevel(logging.WARNING)


def original_logger(name: str):
    """
    Wraps a call to ``logging.getLogger`` to return the original (un-decorated)
    :class:`Logger` object, which should not log any messages to MS Teams or
    Outlook.

    :param name: Name of the logger to retrieve
    :return: the :class:`Logger` object with the un-decorated log methods
    :rtype: OriginalLogAdapter
    """
    logger = getLogger(name)
    return OriginalLogAdapter(logger)


def original_log_method(logger_cls=Logger, log_func_name='_log'):
    """
    Get the original (e.g. un-decorated) log method
    """
    key = f'{logger_cls.__name__}.{log_func_name}'

    if key not in _ORIGINAL_LOG_METHODS:
        # Cache the original log method
        log_method = getattr(logger_cls, log_func_name)
        _ORIGINAL_LOG_METHODS[key] = log_method

    return _ORIGINAL_LOG_METHODS[key]


class OriginalLogAdapter(LoggerAdapter):

    def __init__(self, logger, extra=None):
        super().__init__(logger, extra or {})

    def log(self, level, msg, /, *args, **kwargs):
        if self.isEnabledFor(level):
            self._orig_log(level, msg, args, **kwargs)

    @property
    def _orig_log(self):
        """
        Returns the original :meth:`Logger._log` method for this :attr:`logger`.

        The method resolution is needed at run-time (i.e. when a message is
        logged) in case the :meth:`Logger._log` method is decorated after the
        :class:`OriginalLogAdapter` object is instantiated.

        :rtype: typing.Callable
        """
        return functools.partial(original_log_method(), self.logger)


# Setup logger for the library
LOG = original_logger('aws_teams_logger')



__all__ = ['Globals', 'set_account_name']

from dataclasses import dataclass
from typing import Optional


_EMPTY_VALUES = (None, '')


@dataclass
class __Globals:
    """
    Global defaults
    """

    # Minimum log level for messages sent to Teams
    enabled_lvl: str = 'WARNING'

    # Outbound SES identity
    ses_identity: Optional[str] = None

    # MS Teams email
    teams_email: Optional[str] = None

    # Optional dev emails to notify, when a lambda fails
    dev_emails: Optional[str] = None

    # AWS Account Name (alias)
    account_name: Optional[str] = None

    def __setattr__(self, key, value):
        if value not in _EMPTY_VALUES:
            super().__setattr__(key, value)

    def reset(self):
        """Resets all attributes in the instance."""
        new_obj = self.__class__()
        self.__dict__ = new_obj.__dict__


# Define a singleton instance so we can go through our custom `__setattr__`
Globals = __Globals()


def set_account_name(account_name: str):
    """Convenience function to set the AWS account alias (name)"""
    Globals.account_name = account_name



import inspect
import os
from typing import Dict, Any, List, Tuple, Optional

from .base_logger import _BaseLogger
from .globals import Globals
from ..constants import SOURCE_CODE
from ..log import LOG
from ..models import LambdaContext


class LambdaLogger(_BaseLogger):
    """
    This class can be used to decorate a handler for a Python lambda function.

    For decorating multiple lambda handlers, it is useful to use the method
    `decorate_all_functions`.

    See the documentation in the base class (:class:`_BaseLogger`) for more
    info.

    """
    SUBJECT_PREFIX = 'LAMBDA FAILURE'

    FUNCTION_TYPE = 'Lambda'

    _decorated = []

    def decorate_all_functions(self, enabled_lvl: Optional[str] = None,
                               ses_identity: Optional[str] = None,
                               teams_email: Optional[str] = None,
                               dev_emails: Optional[str] = None):
        """
        Decorates all functions (assumed to be lambda functions) in the calling
        module with the wrapper provided by :class:`LambdaLogger`

        See documentation under the constructor for :class:`_BaseLogger`
        (e.g. `BaseLogger()`) for info on parameters.

        This method call is ideally made after all desired functions to be
        decorated are implemented in the caller module. See below for an
        example.


        module_a.py:

            def f1():   # This function will be decorated
                ...

            def f2():   # This is also decorated
                ...

            LambdaLogger().decorate_all_functions(teams_email='abc123.my.domain@amer.teams.ms')

            def f3():   # This function won't be decorated as it's defined later
                ...

        """
        # Copy over global defaults
        Globals.enabled_lvl = enabled_lvl
        Globals.ses_identity = ses_identity
        Globals.teams_email = teams_email
        Globals.dev_emails = dev_emails

        caller = inspect.stack()[1]
        caller_module_locals = caller[0].f_locals
        caller_module_name = caller_module_locals['__name__']

        # Decorate all functions in module
        decorated_functions = []
        for attr, fn in caller_module_locals.items():
            if self._should_decorate_fn(caller_module_name, fn):
                # Decorate the function if its local to the caller module
                caller_module_locals[attr] = self.__call__(fn)
                decorated_functions.append(attr)

        LOG.debug(
            'Successfully decorated %d lambda functions: %s',
            len(decorated_functions), decorated_functions)

    @classmethod
    def _should_decorate_fn(cls, module_name: str, fn: Any) -> bool:
        """
        Confirm that a function is defined in the module and has not been
        previously decorated.

        Return a boolean indicating whether the function should be decorated.
        """
        if not callable(fn) or fn.__module__ != module_name:
            return False

        key = f'{module_name}.{fn.__name__}'

        if key in cls._decorated:
            return False

        cls._decorated.append(key)
        return True

    def _set_context(self, func, *args, **kwargs):
        try:
            # Context is generally passed as 2nd argument for lambda functions,
            # but to be safe we use the last positional argument instead.
            self.context = args[-1]
        except IndexError:
            # Not a regular lambda function - perhaps a local function used
            # for testing purposes.
            self.context: Any = LambdaContext(func.__name__)

    def _get_context_and_links(self) -> Tuple[Dict[str, Any], List[Dict[str, str]]]:
        links = []
        aws_root = 'https://console.aws.amazon.com'
        aws_region = os.getenv('AWS_REGION', 'us-east-1')

        if isinstance(self.context, LambdaContext):
            # Mock lambda context (lambda function has an empty or missing
            # context argument)
            return {'Function Name': self.context.function_name}, links

        log_group_name = self.context.log_group_name
        log_stream_name = self.context.log_stream_name

        account_id = self.context.invoked_function_arn.split(':')[4]
        account_name = self._get_account_name()

        context = {'Function Name': self.context.function_name,
                   'Request Id': self.context.aws_request_id,
                   'Account Name': account_name,
                   'Account Id': account_id}

        if SOURCE_CODE:
            links.append({'location': SOURCE_CODE, 'text': 'Link to Source'})

        links.append({
            'location': f'{aws_root}/lambda/home?region={aws_region}#/'
                        f'functions/{self.context.function_name}',
            'text': 'Link to Lambda'
        })

        links.append({
            'location': f'{aws_root}/cloudwatch/home?region={aws_region}#logEventViewer:'
                        f'group={log_group_name};stream={log_stream_name}',
            'text': 'Link to Logs'
        })

        return context, links



"""
Utilities for interacting with dates and times.
"""
from datetime import datetime, timezone

import pytz

from ..constants import LOCAL_TZ


# UTC timezone (no offset)
UTC = timezone.utc

# Local timezone object
#
# Adding this globally, in case `LOCAL_TZ` is set as an invalid time zone.
local = pytz.timezone(LOCAL_TZ)


def local_now() -> datetime:
    """
    Returns local time now (by default in Eastern Time)
    """
    return datetime.now(local)


def utc_now() -> datetime:
    """
    Return time in UTC
    """
    return datetime.utcnow().replace(tzinfo=UTC)



import inspect
import os
from json import dumps
from logging import basicConfig
from typing import Optional, List

from aws_teams_logger.log import LOG


template_file = 'template.py'
exclude_dirs = ('__pycache__', )

module_name = os.path.splitext(template_file)[0]


def setup_logging():
    basicConfig(level='INFO')


def get_templates() -> List['TemplateBuilder']:
    """Return all SES templates within the current directory"""
    import importlib

    templates = []
    script_dir = os.path.abspath(os.path.dirname(__file__))
    split_pos = 5
    base_package = '.'.join(script_dir.rsplit('/', split_pos)[-split_pos:])

    tree = os.walk(script_dir, topdown=True)

    for root, dirs, files in tree:
        for dir in exclude_dirs:
            dir in dirs and dirs.remove(dir)

        if root == script_dir:
            continue

        if template_file not in files:
            continue

        package_name = root.split('/')[-1].replace('/', '.')
        module_path = f'{base_package}.{package_name}.{module_name}'

        try:
            template = importlib.import_module(module_path)
        except ModuleNotFoundError:
            LOG.warning(f"Unable to import '{module_path}'")
            continue

        template_data: TemplateBuilder = getattr(template, 'data')
        templates.append(template_data)

    return templates


class TemplateBuilder:
    """
    Class to construct the necessary parameters used to update
    an SES template.

    """
    def __init__(self, name: str, subject: str,
                 html_body: Optional[str] = None):
        self.name = name
        self.subject = subject
        if html_body:
            self.html_contents = html_body
        else:
            caller_dir = os.path.dirname((inspect.stack()[1])[1])
            self.html_contents = self.read_html_contents(caller_dir)

    @staticmethod
    def read_html_contents(path: str, html_file_path='body.html'):
        html_file = os.path.join(path, html_file_path)
        return open(html_file).read()

    def dict(self):
        template_data = {
            "TemplateName": self.name,
            "SubjectPart": self.subject,
            "HtmlPart": self.html_contents
        }

        return template_data

    def __str__(self):
        return dumps(self.dict())



"""
Models for the 'Logger' decorator classes

"""
__all__ = ['LambdaContext', 'TaskContext', '_BaseContext']

import os
from dataclasses import dataclass
from typing import Callable, Dict, Optional, Any

import requests

from .log import LOG


@dataclass
class _BaseContext:
    """
    Base Context class, which defines the minimum attributes that a Context
    object would be expected to contain.
    """
    function_name: str


class TaskContext(_BaseContext):

    def __init__(self, fn: Callable):
        """
        Create a :class:`TaskContext` object.
        """
        super().__init__(fn.__qualname__)

    @classmethod
    def get_ecs_metadata(cls, endpoint_env_var='ECS_CONTAINER_METADATA_URI_V4'
                         ) -> Optional[Dict[str, Any]]:
        """
        Retrieve metadata for the ECS or Fargate task, if available.
        Uses the Task Metadata V4 endpoint.

        Ref: https://docs.aws.amazon.com/AmazonECS/latest/userguide/task-metadata-endpoint-v4-fargate.html
        """
        metadata_url = os.getenv(endpoint_env_var)

        if not metadata_url:
            docs_link = ('https://docs.aws.amazon.com/AmazonECS/latest/userguide/'
                         'task-metadata-endpoint-v4-fargate.html')
            LOG.info(
                f'Environment variable "{endpoint_env_var}" not defined '
                'in task; consider updating to platform version 1.4.0 to enable '
                'this feature. Please refer to the following docs:\n'
                f'  {docs_link}')
            return

        return cls._retrieve_ecs_metadata(metadata_url)

    @staticmethod
    def _retrieve_ecs_metadata(metadata_url: str) -> Dict[str, Any]:
        """
        Make a call to the ECS Metadata endpoint and return the result as a
        dictionary object.
        """
        r = requests.get(metadata_url)
        r.raise_for_status()

        return r.json()


class LambdaContext(_BaseContext):

    def __init__(self, function_name: str,
                 request_id=None, account_id=None, account_name=None):
        """
        Create a :class:`LambdaContext` object.
        """
        super().__init__(function_name)

        self.request_id = request_id
        self.account_id = account_id
        self.account_name = account_name



import functools
import logging
import logging.config
import sys
import traceback
from io import StringIO
from typing import Optional, Tuple, Dict, Any, List, Callable, Union

import boto3
from botocore.exceptions import ClientError

from .globals import Globals
from ..constants import *
from ..log import LOG, setup_logging, original_log_method
from ..models import _BaseContext
from ..utils.aws.ses import SESHelper
from ..utils.date_util import local_now
from ..utils.decorators import log_time
from ..utils.types import get_formatted_message


class _BaseLogger:

    # Prefix string for the 'subject' used in email messages
    SUBJECT_PREFIX = 'FAILURE'

    # The type or classification of the decorated function
    #
    # For example, 'Lambda' indicates an AWS Lambda function.
    FUNCTION_TYPE = 'Base'

    # Checks whether required vars are validated (either defined in environment
    # or passed in as parameters). This should only be run once for all
    # decorated functions.
    #
    # Note: don't directly modify this value; use `reset_validation_state` to
    # reset this attribute if needed.
    _VALIDATED = False

    def __init__(self, f: Optional[Callable] = None, *,
                 enabled_lvl: Optional[str] = None,
                 ses_identity: Optional[str] = None,
                 teams_email: Optional[str] = None,
                 dev_emails: Optional[str] = None,
                 logger_cls=logging.Logger,
                 log_func_name='_log',
                 raise_=True):
        """
        Create and return a :class:`_BaseLogger` object.

        This defines a decorator method which handles the posting of warning
        level or above log messages (used via the builtin ``logging`` module)
        to a specified Microsoft Teams channel. Log messages with an `exc_info`,
        as well as any uncaught exceptions, will also be sent to an optional
        list of Developer Emails via Microsoft Outlook.

        This is an abstract class and should not be used directly. Instead,
        please import the concrete sub-classes which implement all required
        methods in this containing class:

            * :class:`LambdaLogger` - The main use case is to decorate a handler for
              a Python lambda function. For decorating multiple lambda handlers,
              it is useful to invoke the helper method `decorate_all_functions`.

            * :class:`TaskLogger` - Intended to be used with ECS functions, such as
              functions that will be run in a Fargate task. However this can also
              be used to decorate any generic functions as well.

        The following environment variables will be used if provided:

            * `AWS_ACCOUNT_NAME` - AWS Account Alias (name); if defined, will
              be used instead of making a call to `iam:ListAccountAliases`
              to retrieve the alias of the current AWS account.

            * `TEAMS_LOG_LVL` - Can be used instead of the ``enabled_lvl``
              parameter. This determines the minimum log level for messages
              sent to Teams. The default value is 'WARNING', which means
              that log messages at the 'WARNING', 'ERROR', and 'EXCEPTION'
              levels will be sent to Teams, but not messages at the 'INFO'
              or 'DEBUG' level.

            * `TEAMS_EMAIL` - The email to send log messages or lambda failures
              (e.g. uncaught exceptions) to. Required, if not passed in via the
              constructor.

            * `SES_IDENTITY` - Sender or outbound email, must be validated under
              SES in the AWS console. Required, if not passed in via the
              constructor.

            * `DEV_EMAILS` - Comma delimited list of dev emails, if provided
              will send stylized HTML to them when any uncaught exceptions are
              raised. Example: 'user1@my.domain.org,user2@my.domain.org'

            * `SOURCE_CODE` - A link to the source code repo (such as Bitbucket)
               for the project.

            * `AWS_LOG_GROUP` - Only applies when the `TaskLogger` decorator is
              used. Determines the log group to link to in the AWS console.
              Generally this is not needed to be specified (see note on 'ECS
              Tasks' below)

            * `LOG_LEVEL` - The minimum log level for messages logged by this
              library; only determines if they show up in CloudWatch
              (Example: 'DEBUG')

            * `LOCAL_TZ` - User's local time zone, passed in to ``pytz.timezone``;
              used when generating the date/time in the subject for a Teams or
              Devs email message (Example: 'US/Eastern')

        The `logger_cls` and `log_func_name` together determine the base log
        function to decorate. By default we decorate the `logging.Logger._log`
        method, which is the base method used by all logger methods.

        The constructor also accepts an optional `enabled_lvl` parameter, which
        determines the minimum log level (ex. "ERROR") at which messages are
        logged to Teams, but the preferred option is to globally set the
        environment variable 'TEAMS_LOG_LVL', as this will also work better
        when the decorating multiple lambda handlers in a module.


        Simple Usage:

            >>> from aws_teams_logger import LambdaLogger

            >>> log = logging.getLogger(__name__)

            >>> @LambdaLogger
            >>> def my_lambda_handler(event: Dict[str, Any], context: Any):
                >>> ...
                >>> other_func()
                >>> ...

            >>> def other_func():
                # Message is not logged to Teams (default log level is "WARN")
                >>> log.info('Info level log')
                # This message will be forwarded to Teams
                >>> log.warning('Sample Warn message')
                # Messages with the `exc_info` parameter will be logged to both
                # the Teams Channel and any subscribed Dev Emails via Outlook
                >>> try:
                    >>> empty_dict = {}
                    >>> value = empty_dict['missing key']
                >>> except KeyError:
                    >>> log.error('Key missing from `empty_dict`', exc_info=True)
                # Uncaught errors will be logged to both Teams and any Dev Emails
                >>> finally:
                    >>> result = 1 / 0

        For ECS Tasks:

            The decorator retrieves ECS metadata on the currently running task
            when posting messages to Teams or via email. It will automatically
            retrieve data from the Task Metadata V4 endpoint, if this is
            available. To enable correct data reporting for the the V4 endpoint,
            set the `Platform Version` for the task to "1.4.0" or higher as
            mentioned in the article below.

            https://docs.aws.amazon.com/AmazonECS/latest/userguide/task-metadata-endpoint-v4-fargate.html


            Usage for Tasks:

                >>> from aws_teams_logger import TaskLogger

                >>> log = logging.getLogger(__name__)

                >>> class MyTaskClass:
                    >>>
                    >>> @classmethod
                    >>> @TaskLogger
                    >>> def my_task_func(cls, *args, **kwargs):
                        >>> ...
                        >>> cls.other_func()
                        >>> ...
                    >>>
                    >>> @staticmethod
                    >>> def other_func():
                        # See logging example under "Simple Usage" section above
                        >>> ...

        """
        # Copy over global defaults
        Globals.ses_identity = ses_identity
        Globals.teams_email = teams_email
        Globals.dev_emails = dev_emails
        # Setup logging config
        setup_logging()

        self._func = f
        functools.update_wrapper(self, self._func)

        self.enabled_lvl = self._get_enabled_lvl(enabled_lvl)
        self.logger_cls = logger_cls
        self.log_func_name = log_func_name
        self.raise_ = raise_
        # noinspection PyTypeChecker
        # This property will be set by the sub-classes.
        self.context: _BaseContext = None

    def __call__(self, *param_args, **param_kwargs):
        """
        Decorates (wraps) function `f` to to stream log messages at the
        `enabled_lvl` or above, as well as any uncaught exceptions, to the
        specified Teams channel.

        Also optionally sends any errors to a comma-separated list of
        'DEVS_EMAIL', if this value is defined.

        """
        # Get original (e.g. un-decorated) log method
        log_method = original_log_method(self.logger_cls, self.log_func_name)
        # Validate any required vars here, so we don't run into errors later
        # when the emails need to be sent.
        self._validate_vars_if_needed()

        def decorator(func):
            @functools.wraps(func)
            def new_func(*args, **kwargs):

                # Set context object
                self._set_context(func, *args, **kwargs)

                # Decorate log method - needs to be inside the decorator to avoid
                # potential bugs when multiple decorated functions are run.
                setattr(self.logger_cls, self.log_func_name,
                        self._decorate_log_method(log_method))

                try:
                    return func(*args, **kwargs)

                except Exception as e:
                    if hasattr(e, 'code') and hasattr(e, 'message'):
                        # Exceptions with :attr:`code` and :attr:`message` in our case
                        # should already be logged to Teams
                        LOG.error('Failure, err_code=%s, err_msg=%s',
                                  e.code, e.message, exc_info=e)

                    else:
                        LOG.error('Failure: %s', str(e), exc_info=e)
                        # Send lambda failure to Teams
                        self._send_to_ses(error=self._format_exception(e))

                    if self.raise_:
                        raise

            return new_func

        # We're called without parens - ex. @DecoratorClass
        if self._func:
            return decorator(self._func)(*param_args, **param_kwargs)

        # We're called with parens - ex. @DecoratorClass(key='value')
        # In this case, the only argument will be the function to decorate.
        try:
            func = param_args[0]
        except IndexError:
            # Normally this won't happen, if the decorator is used properly
            return decorator

        return decorator(func)

    @property
    def ses(self):
        return SESHelper()

    @property
    def ses_identity(self):
        return SES_IDENTITY or Globals.ses_identity

    @property
    def teams_email(self):
        return TEAMS_EMAIL or Globals.teams_email

    @property
    def dev_emails(self):
        return DEV_EMAILS or Globals.dev_emails

    @staticmethod
    def reset_validation_state():
        """
        Used to clear :attr:`_VALIDATED` in case we need to re-validate when
        the :meth:`__call__` decorator is run again.
        """
        _BaseLogger._VALIDATED = False

    @classmethod
    def _get_account_name(cls):
        """
        Retrieve the human-readable account name to show in logs for a given
        AWS Account ID.

        If an env variable 'AWS_ACCOUNT_NAME' is defined, this value will
        be used instead.

        Note: Lambdas and ECS tasks must otherwise have the necessary
        permissions (shown below) for this call to be successful. If not,
        the account name won't be resolved and an error will be logged to
        CloudWatch.

            {
                "Effect": "Allow",
                "Action": "iam:ListAccountAliases",
                "Resource": "*"
            }

        """

        def get_account_alias(default_alias='Unknown') -> str:
            """Retrieve account name from environment, or IAM if undefined."""

            # Short path: check if the account alias is defined in environment
            if AWS_ACCOUNT_NAME:
                return AWS_ACCOUNT_NAME

            # Retrieve the alias from an `iam:ListAccountAliases` API call
            try:
                @log_time(log_message='Retrieved the AWS account alias from IAM')
                def get_account_alias_from_iam():
                    return boto3.client('iam').list_account_aliases()['AccountAliases'][0]

                return get_account_alias_from_iam()

            except Exception as e:
                LOG.error('Unable to retrieve the account alias, please ensure '
                          'the attached role has the necessary permissions '
                          '(iam:ListAccountAliases). Error: %s', e)

                return default_alias

        if not Globals.account_name:
            Globals.account_name = get_account_alias()

        return Globals.account_name

    def _set_context(self, func, *args, **kwargs):
        """
        Set the context object, for example a Lambda Context object.

        An implementation should set the :attr:`context` object which can extend
        from the :class:`_BaseContext` class.
        """
        raise NotImplementedError(
            'Sub-classes should implement this method')

    def _get_context_and_links(self) -> Tuple[Dict[str, Any], List[Dict[str, str]]]:
        """
        Return a two-tuple of (context_data, links)

        `context_data` is a dictionary containing key:value pairs to show in a
        Teams message, and `links` is a list of dictionary objects each
        containing a "location" and "text" field.
        """
        raise NotImplementedError(
            'Sub-classes should implement this method')

    def _get_subject(self, dt_format='%m/%d/%Y %I:%M%p') -> str:
        """
        Generate a subject for the Teams or Devs email message.

        Uses the class attribute `SUBJECT_PREFIX` which can optionally be set.
        """
        dt_now = local_now().strftime(dt_format)
        return f'{self.SUBJECT_PREFIX}: {self.context.function_name}, {dt_now}'

    def _validate_vars_if_needed(self):
        if _BaseLogger._VALIDATED:
            return

        if not self.ses_identity:
            raise ValueError(
                'SES Identity must be set, either as an argument to the '
                'constructor or via the "SES_IDENTITY" environment variable.')

        if not self.teams_email:
            raise ValueError(
                'MS Teams Email (ex: abc123.my.domain@amer.teams.ms) must be '
                'set, either as an argument to the constructor or via the '
                '"TEAMS_EMAIL" environment variable.')

        _BaseLogger._VALIDATED = True

    @staticmethod
    def _get_enabled_lvl(enabled_lvl: Optional[str] = None) -> int:
        """
        Get enabled log level, e.g. minimum log level to stream messages to Teams
        """
        default_enabled_log_lvl = enabled_lvl or Globals.enabled_lvl
        return logging._nameToLevel[TEAMS_LOG_LVL or default_enabled_log_lvl]

    def _decorate_log_method(self, f):
        """
        Decorate (wrap) the specified base log method.

        By default, decorates the base method func:`Logger._log` from the
        `logging` library.
        """
        @functools.wraps(f)
        def new_log_func(self_: logging.Logger, level, msg, msg_args,
                         exc_info=None, *args, **kwargs):

            if level >= self.enabled_lvl:
                send_kwargs = {'msg': get_formatted_message(msg, msg_args),
                               'lvl': level}
                if exc_info:
                    send_kwargs['error'] = self._format_exception(exc_info)

                self._send_to_ses(**send_kwargs)

            return f(self_, level, msg, msg_args,
                     exc_info=exc_info, *args, **kwargs)

        return new_log_func

    @staticmethod
    def _format_exception(exc_info):
        """
        Format `exc_info` or current exception info as a tuple of
        (type, value, traceback).

        Copied from the ``logging`` module for now.

        """
        if isinstance(exc_info, BaseException):
            return type(exc_info), exc_info, exc_info.__traceback__
        elif not isinstance(exc_info, tuple):
            return sys.exc_info()
        return exc_info

    @classmethod
    def _get_msg_with_exc_trace(cls, exc_info, msg: Optional[str] = None):
        """
        Get error message with the error traceback included.
        Copied directly from the ``logging`` module for now.

        TODO: Perhaps update to decorate `Logger.handle` method instead,
            that way we can get the formatted `exc_info` directly.
        """
        def _format_exc_traceback(ei) -> str:
            sio = StringIO()
            tb = ei[2]
            traceback.print_exception(ei[0], ei[1], tb, None, sio)
            s = sio.getvalue()
            sio.close()
            if s[-1:] == '\n':
                s = s[:-1]
            return s

        if exc_info:
            exc_info = cls._format_exception(exc_info)
            exc_text = _format_exc_traceback(exc_info)
            if exc_text:
                return f'{msg.strip()}\n\n{exc_text}' if msg else exc_text

        return msg

    def _send_to_ses(self, *, msg: Optional[str] = None, lvl: Optional[str] = None,
                     error=None):
        """
        Forwards a log message `msg` logged at `lvl` or an exception `error` to
        the Teams channel

        This method accepts keyword arguments only, to avoid any confusion.

        If uncaught exceptions were raised in the decorated handler and env
        variable `DEV_EMAILS` is provided, send out a specially formatted email
        to those users as well.

        """
        subject = self._get_subject()
        context, links = self._get_context_and_links()

        send_data = {'subject': subject,
                     'context': context,
                     'message': msg,
                     'links': links}
        if error:
            self._notify_dev_emails()

            error_msg_with_tb = self._get_msg_with_exc_trace(error, msg)

            if lvl:
                lvl_name = logging._levelToName.get(lvl, logging.ERROR)
                error_msg_with_tb = f'[{lvl_name.upper()}] {error_msg_with_tb}'
                send_data['level'] = lvl_name.capitalize()

            send_data['error'] = {
                'class': error[0].__name__,
                'message': getattr(error[1], 'message', error_msg_with_tb)
            }

        self._send_templated_email('send-to-teams', send_data, self.teams_email)

    def _notify_dev_emails(self):
        """
        Notifies any devs about an error, if needed
        """
        if not self.dev_emails:
            return

        subject = self._get_subject()
        context, links = self._get_context_and_links()

        context_data = context.copy()

        # Retrieve account info
        account_data = {}
        for key in 'Account Name', 'Account Id':
            value = context_data.pop(key, None)
            if value:
                account_data[key] = value

        # Commenting this out for now - would be nice to use, but
        # it's specific to Python 3.8 and above.
        # account_data = {k: v for k in ('Account Name', 'Account Id')
        #                 if (v := context_data.pop(k, None))}

        send_outlook_data = {'subject': subject,
                             'function_type': self.FUNCTION_TYPE,
                             'context': context_data,
                             'account': account_data,
                             'links': links}

        self._send_templated_email(
            'send-to-outlook', send_outlook_data, self.dev_emails)

    def _send_templated_email(self, name: str,
                              data: Dict[str, Any],
                              to_addresses: Union[List[str], str]):
        """
        Sends a templated email using SES. If an error occurs, log a message if
        the error is a known one, otherwise raise the original error.

        :param name: Name of the SES template
        :param data: Template data as a dictionary object
        :param to_addresses: List of recipients for the email
        """
        try:
            self.ses.send_templated_email(name, data, to_addresses,
                                          self.ses_identity, self.ses_identity)

        except ClientError as ce:
            error = ce.response.get('Error', {})
            error_code = error.get('Code', 'Unknown')

            if error_code == 'TemplateDoesNotExist':
                LOG.error(
                    f'Template {name} does not exist; please call '
                    f'`upload_templates` to upload the required SES templates.')
            else:
                raise



__all__ = ['as_list',
           'get_formatted_message']

import collections.abc
from typing import Union, List


def as_list(o: Union[str, List[str]], sep=','):
    """
    Return `o` if already a list. If `o` is None or an empty string,
    return an empty list. Otherwise, split the string on `sep` and
    return the list result.

    """
    if not o:
        return []

    if isinstance(o, list):
        return o

    return o.split(sep)


def get_formatted_message(msg, args):
    """
    Gets a message after formatting is applied. Ideally, this would be the arguments
    passed to a logger method, such as `LOG.warning`.

    """
    if (args and len(args) == 1 and isinstance(args[0], collections.abc.Mapping)
            and args[0]):
        args = args[0]
    msg = str(msg)
    if args:
        msg = msg % args
    return msg



from aws_teams_logger.log import LOG
from .utils import get_templates, setup_logging
from ..ses import SESHelper


def upload_templates(profile_name=None):
    """
    Upload templates to SES, updating any existing templates if necessary.

    Walks through all packages under the current directory. If the package
    contains a :attr:`template_file` which defines a `data` attribute
    representing the template data, it replaces the template on SES with
    the current version on the local.

    """
    setup_logging()
    ses = SESHelper(profile_name=profile_name)

    for template in get_templates():

        LOG.info('[ %s ]', template.name)
        LOG.info(f'Uploading SES template...')
        success = ses.replace_template(template.dict())
        if success:
            LOG.info('  SUCCESS')
        else:
            LOG.warning('  FAILED')



import os
from typing import Tuple, Dict, Any, List

from .base_logger import _BaseLogger
from ..constants import AWS_LOG_GROUP, SOURCE_CODE
from ..log import LOG
from ..models import TaskContext


class TaskLogger(_BaseLogger):
    """
    This class can be used to decorate ECS functions or methods, such as ones
    that will be run in a Fargate task. However this can also be used to
    decorate any generic functions or methods as well.

    See the documentation in the base class (:class:`_BaseLogger`) for more
    info.

    """
    SUBJECT_PREFIX = 'TASK FAILURE'

    FUNCTION_TYPE = 'Task'

    _task_metadata = None

    @classmethod
    def _set_task_metadata(cls):
        if cls._task_metadata is None:
            LOG.info('Retrieving ECS task metadata')
            cls._task_metadata = TaskContext.get_ecs_metadata()

    def _set_context(self, func, *args, **kwargs):
        self.context = TaskContext(func)
        self._set_task_metadata()

    def _get_context_and_links(self) -> Tuple[Dict[str, Any], List[Dict[str, str]]]:

        context, links = {}, []
        aws_root = 'https://console.aws.amazon.com'
        aws_region = os.getenv('AWS_REGION', 'us-east-1')

        cluster_name = None
        image_id = image = None
        log_group_name = AWS_LOG_GROUP
        logs_region_name = aws_region
        log_stream_name = None

        if self._task_metadata:
            container_labels = self._task_metadata.get('Labels', {})
            ecs_cluster_info = container_labels.get('com.amazonaws.ecs.cluster')
            log_options = self._task_metadata.get('LogOptions', {})
            image = self._task_metadata.get('Image')
            image_id = self._task_metadata.get('ImageID')

            if image:
                context['ECR Repo'] = image.split('/', 1)[-1].split(':', 1)[0]

            context['Function Name'] = self.context.function_name

            if ecs_cluster_info:
                ecs_cluster_parts = ecs_cluster_info.split(':')
                aws_region = ecs_cluster_parts[3]
                account_id = ecs_cluster_parts[4]
                account_name = self._get_account_name()
                cluster_name = ecs_cluster_parts[-1].split('/', 1)[-1]
                # Add account details to context
                context.update({'Account Name': account_name,
                                'Account Id': account_id})

            if log_options:
                log_group_name = log_options.get('awslogs-group')
                logs_region_name = log_options.get('awslogs-region')
                log_stream_name = log_options.get('awslogs-stream')

        else:
            context['Function Name'] = self.context.function_name

        if SOURCE_CODE:
            links.append({'location': SOURCE_CODE, 'text': 'Link to Source'})

        if cluster_name:
            links.append({
                'location': f'{aws_root}/ecs/home?region={aws_region}#/clusters/{cluster_name}',
                'text': 'Link to ECS Cluster'
            })
        else:
            links.append({
                'location': f'{aws_root}/ecs/home?region={aws_region}#/clusters',
                'text': 'Link to ECS'
            })

        if image and image_id:
            account_id = image.split('.', 1)[0]
            links.append({
                'location': f'{aws_root}/ecr/repositories/private/{account_id}/'
                            f'{context["ECR Repo"]}/image/{image_id}/details/'
                            f'?region={aws_region}',
                'text': 'Link to Image'
            })

        if log_group_name:
            if log_stream_name:
                location = (f'{aws_root}/cloudwatch/home?region={logs_region_name}'
                            f'#logEventViewer:group={log_group_name};stream={log_stream_name}')
            else:
                location = (f'{aws_root}/cloudwatch/home?region={logs_region_name}'
                            f'#logEventViewer:group={log_group_name}')

            links.append({
                'location': location,
                'text': 'Link to Logs'
            })

        return context, links



import pathlib

from setuptools import setup, find_packages


here = pathlib.Path(__file__).parent

package_name = 'aws_teams_logger'

packages = find_packages(exclude=['tests.*', 'tests'])

requires = [
    'pytz>=2018.1,<2022.0',
    'requests'
]

test_requirements = [
    'pytest-cov',
    'pytest-mock',
    'pytest>=6,<7'
]

about = {}
exec((here / package_name / '__version__.py').read_text(), about)

readme = (here / 'README.md').read_text()

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=packages,
    package_data={'': ['LICENSE']},
    package_dir={package_name: package_name},
    include_package_data=True,
    install_requires=requires,
    license=about['__license__'],
    # download_url='https://github.com/user/reponame/archive/v_01.tar.gz',
    keywords=['teams logger', 'ms-teams-logger',
              'microsoft teams', 'ms-teams',
              'teams logger for AWS',
              'aws-lambda', 'ecs', 'fargate'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: System :: Logging',
        'Topic :: Communications :: Chat',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    python_requires='>=3.7',
    tests_require=test_requirements,
    extras_require={
        'standalone': ['boto3~=1.17.85']
    }
)



"""
MS Teams / Email Logger
"""
__title__ = 'aws-teams-logger'
__description__ = 'Python MS Teams and Email logger for AWS environments'
__url__ = 'https://github.com/rnag/aws-teams-logger'
__version__ = '0.5.2'
__author__ = 'Ritvik Nag'
__author_email__ = 'rv.kvetch@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2021 Ritvik Nag'



"""
Time utilities
"""
import sys
import time


# Preferred clock, based on which one is more accurate on a given system.
# Ported from `requests`
if sys.platform == 'win32':
    try:  # Python 3.4+
        preferred_clock = time.perf_counter
    except AttributeError:  # Earlier than Python 3.
        preferred_clock = time.clock
else:
    preferred_clock = time.time
